@@ -1,13 +1,15 @@
 """This package contains the "front end" classes and functions
 for Beaker caching.
 
-Included are the :class:`.Cache` and :class:`.CacheManager` classes, 
+Included are the :class:`.Cache` and :class:`.CacheManager` classes,
 as well as the function decorators :func:`.region_decorate`,
 :func:`.region_invalidate`.
 
 """
 import warnings
+from itertools import chain
 
+from beaker._compat import u_, unicode_text, func_signature, bindfuncargs
 import beaker.container as container
 import beaker.util as util
 from beaker.crypto.util import sha1
@@ -18,13 +20,15 @@ import beaker.ext.memcached as memcached
 import beaker.ext.database as database
 import beaker.ext.sqla as sqla
 import beaker.ext.google as google
-import collections
+import beaker.ext.mongodb as mongodb
+import beaker.ext.redisnm as redisnm
+from functools import wraps
 
 # Initialize the cache region dict
 cache_regions = {}
 """Dictionary of 'region' arguments.
 
-A "region" is a string name that refers to a series of cache 
+A "region" is a string name that refers to a series of cache
 configuration arguments.    An application may have multiple
 "regions" - one which stores things in a memory cache, one
 which writes data to files, etc.
@@ -35,11 +39,11 @@ of configuration arguments.  Example::
     from beaker.cache import cache_regions
     cache_regions.update({
         'short_term':{
-            'expire':'60',
+            'expire':60,
             'type':'memory'
         },
         'long_term':{
-            'expire':'1800',
+            'expire':1800,
             'type':'dbm',
             'data_dir':'/tmp',
         }
@@ -49,6 +53,7 @@ of configuration arguments.  Example::
 
 cache_managers = {}
 
+
 class _backends(object):
     initialized = False
 
@@ -95,28 +100,34 @@ class _backends(object):
                     # Warn when there's a problem loading a NamespaceManager
                     if not isinstance(sys.exc_info()[1], DistributionNotFound):
                         import traceback
-                        from io import StringIO
+                        try:
+                            from StringIO import StringIO  # Python2
+                        except ImportError:
+                            from io import StringIO        # Python3
+
                         tb = StringIO()
                         traceback.print_exc(file=tb)
                         warnings.warn(
                             "Unable to load NamespaceManager "
                             "entry point: '%s': %s" % (
-                                        entry_point, 
-                                        tb.getvalue()), 
+                                        entry_point,
+                                        tb.getvalue()),
                                         RuntimeWarning, 2)
         except ImportError:
             pass
 
 # Initialize the basic available backends
 clsmap = _backends({
-          'memory':container.MemoryNamespaceManager,
-          'dbm':container.DBMNamespaceManager,
-          'file':container.FileNamespaceManager,
-          'ext:memcached':memcached.MemcachedNamespaceManager,
-          'ext:database':database.DatabaseNamespaceManager,
-          'ext:sqla': sqla.SqlaNamespaceManager,
-          'ext:google': google.GoogleNamespaceManager,
-          })
+    'memory': container.MemoryNamespaceManager,
+    'dbm': container.DBMNamespaceManager,
+    'file': container.FileNamespaceManager,
+    'ext:memcached': memcached.MemcachedNamespaceManager,
+    'ext:database': database.DatabaseNamespaceManager,
+    'ext:sqla': sqla.SqlaNamespaceManager,
+    'ext:google': google.GoogleNamespaceManager,
+    'ext:mongodb': mongodb.MongoNamespaceManager,
+    'ext:redis': redisnm.RedisNamespaceManager
+})
 
 
 def cache_region(region, *args):
@@ -126,11 +137,11 @@ def cache_region(region, *args):
     Example::
 
         from beaker.cache import cache_regions, cache_region
-        
+
         # configure regions
         cache_regions.update({
             'short_term':{
-                'expire':'60',
+                'expire':60,
                 'type':'memory'
             }
         })
@@ -139,10 +150,10 @@ def cache_region(region, *args):
         def load(search_term, limit, offset):
             '''Load from a database given a search term, limit, offset.'''
             return database.query(search_term)[offset:offset + limit]
-    
+
     The decorator can also be used with object methods.  The ``self``
-    argument is not part of the cache key.  This is based on the 
-    actual string name ``self`` being in the first argument 
+    argument is not part of the cache key.  This is based on the
+    actual string name ``self`` being in the first argument
     position (new in 1.6)::
 
         class MyThing(object):
@@ -150,27 +161,27 @@ def cache_region(region, *args):
             def load(self, search_term, limit, offset):
                 '''Load from a database given a search term, limit, offset.'''
                 return database.query(search_term)[offset:offset + limit]
-    
+
     Classmethods work as well - use ``cls`` as the name of the class argument,
     and place the decorator around the function underneath ``@classmethod``
     (new in 1.6)::
-    
+
         class MyThing(object):
             @classmethod
             @cache_region('short_term', 'load_things')
             def load(cls, search_term, limit, offset):
                 '''Load from a database given a search term, limit, offset.'''
                 return database.query(search_term)[offset:offset + limit]
-    
+
     :param region: String name of the region corresponding to the desired
       caching arguments, established in :attr:`.cache_regions`.
-      
+
     :param \*args: Optional ``str()``-compatible arguments which will uniquely
       identify the key used by this decorated function, in addition
       to the positional arguments passed to the function itself at call time.
       This is recommended as it is needed to distinguish between any two functions
       or methods that have the same name (regardless of parent class or not).
-      
+
     .. note::
 
         The function being decorated must only be called with
@@ -181,14 +192,15 @@ def cache_region(region, *args):
         forms the unique cache key.
 
     .. note::
-    
+
         When a method on a class is decorated, the ``self`` or ``cls``
         argument in the first position is
         not included in the "key" used for caching.   New in 1.6.
-        
+
     """
     return _cache_decorate(args, None, None, region)
 
+
 def region_invalidate(namespace, region, *args):
     """Invalidate a cache region corresponding to a function
     decorated with :func:`.cache_region`.
@@ -210,11 +222,11 @@ def region_invalidate(namespace, region, *args):
     Example::
 
         from beaker.cache import cache_regions, cache_region, region_invalidate
-        
+
         # configure regions
         cache_regions.update({
             'short_term':{
-                'expire':'60',
+                'expire':60,
                 'type':'memory'
             }
         })
@@ -227,11 +239,11 @@ def region_invalidate(namespace, region, *args):
         def invalidate_search(search_term, limit, offset):
             '''Invalidate the cached storage for a given search term, limit, offset.'''
             region_invalidate(load, 'short_term', 'load_data', search_term, limit, offset)
-    
+
     Note that when a method on a class is decorated, the first argument ``cls``
     or ``self`` is not included in the cache key.  This means you don't send
     it to :func:`.region_invalidate`::
-    
+
         class MyThing(object):
             @cache_region('short_term', 'some_data')
             def load(self, search_term, limit, offset):
@@ -241,9 +253,9 @@ def region_invalidate(namespace, region, *args):
             def invalidate_search(self, search_term, limit, offset):
                 '''Invalidate the cached storage for a given search term, limit, offset.'''
                 region_invalidate(self.load, 'short_term', 'some_data', search_term, limit, offset)
-        
+
     """
-    if isinstance(namespace, collections.Callable):
+    if callable(namespace):
         if not region:
             region = namespace._arg_region
         namespace = namespace._arg_namespace
@@ -255,7 +267,9 @@ def region_invalidate(namespace, region, *args):
         region = cache_regions[region]
 
     cache = Cache._get_cache(namespace, region)
-    _cache_decorator_invalidate(cache, region['key_length'], args)
+    _cache_decorator_invalidate(cache,
+                                region.get('key_length', util.DEFAULT_CACHE_KEY_LENGTH),
+                                args)
 
 
 class Cache(object):
@@ -280,6 +294,10 @@ class Cache(object):
                 raise cls
         except KeyError:
             raise TypeError("Unknown cache implementation %r" % type)
+
+        if expire is not None:
+            expire = int(expire)
+
         self.namespace_name = namespace
         self.namespace = cls(namespace, **nsargs)
         self.expiretime = expiretime or expire
@@ -310,7 +328,7 @@ class Cache(object):
     remove = remove_value
 
     def _get_value(self, key, **kw):
-        if isinstance(key, str):
+        if isinstance(key, unicode_text):
             key = key.encode('ascii', 'backslashreplace')
 
         if 'type' in kw:
@@ -332,7 +350,7 @@ class Cache(object):
         kwargs = self.nsargs.copy()
         kwargs.update(kw)
         c = Cache(self.namespace.namespace, type=type, **kwargs)
-        return c._get_value(key, expiretime=expiretime, createfunc=createfunc, 
+        return c._get_value(key, expiretime=expiretime, createfunc=createfunc,
                             starttime=starttime)
 
     def clear(self):
@@ -475,7 +493,7 @@ class CacheManager(object):
         .. note::
 
             The function being decorated must only be called with
-            positional arguments. 
+            positional arguments.
 
         """
         return _cache_decorate(args, self, kwargs, None)
@@ -517,12 +535,14 @@ class CacheManager(object):
 
         cache = self.get_cache(namespace, **kwargs)
         if hasattr(func, '_arg_region'):
-            key_length = cache_regions[func._arg_region]['key_length']
+            cachereg = cache_regions[func._arg_region]
+            key_length = cachereg.get('key_length', util.DEFAULT_CACHE_KEY_LENGTH)
         else:
-            key_length = kwargs.pop('key_length', 250)
+            key_length = kwargs.pop('key_length', util.DEFAULT_CACHE_KEY_LENGTH)
         _cache_decorator_invalidate(cache, key_length, args)
 
-def _cache_decorate(deco_args, manager, kwargs, region):
+
+def _cache_decorate(deco_args, manager, options, region):
     """Return a caching function decorator."""
 
     cache = [None]
@@ -530,7 +550,10 @@ def _cache_decorate(deco_args, manager, kwargs, region):
     def decorate(func):
         namespace = util.func_namespace(func)
         skip_self = util.has_self_arg(func)
-        def cached(*args):
+        signature = func_signature(func)
+
+        @wraps(func)
+        def cached(*args, **kwargs):
             if not cache[0]:
                 if region is not None:
                     if region not in cache_regions:
@@ -538,32 +561,42 @@ def _cache_decorate(deco_args, manager, kwargs, region):
                             'Cache region not configured: %s' % region)
                     reg = cache_regions[region]
                     if not reg.get('enabled', True):
-                        return func(*args)
+                        return func(*args, **kwargs)
                     cache[0] = Cache._get_cache(namespace, reg)
                 elif manager:
-                    cache[0] = manager.get_cache(namespace, **kwargs)
+                    cache[0] = manager.get_cache(namespace, **options)
                 else:
                     raise Exception("'manager + kwargs' or 'region' "
                                     "argument is required")
 
+            cache_key_kwargs = []
+            if kwargs:
+                # kwargs provided, merge them in positional args
+                # to avoid having different cache keys.
+                args, kwargs = bindfuncargs(signature, args, kwargs)
+                cache_key_kwargs = [u_(':').join((u_(key), u_(value))) for key, value in kwargs.items()]
+
+            cache_key_args = args
             if skip_self:
-                try:
-                    cache_key = " ".join(map(str, deco_args + args[1:]))
-                except UnicodeEncodeError:
-                    cache_key = " ".join(map(str, deco_args + args[1:]))
-            else:
-                try:
-                    cache_key = " ".join(map(str, deco_args + args))
-                except UnicodeEncodeError:
-                    cache_key = " ".join(map(str, deco_args + args))
+                cache_key_args = args[1:]
+
+            cache_key = u_(" ").join(map(u_, chain(deco_args, cache_key_args, cache_key_kwargs)))
+
             if region:
-                key_length = cache_regions[region]['key_length']
+                cachereg = cache_regions[region]
+                key_length = cachereg.get('key_length', util.DEFAULT_CACHE_KEY_LENGTH)
             else:
-                key_length = kwargs.pop('key_length', 250)
-            if len(cache_key) + len(namespace) > key_length:
-                cache_key = sha1(cache_key).hexdigest()
+                key_length = options.pop('key_length', util.DEFAULT_CACHE_KEY_LENGTH)
+
+            # TODO: This is probably a bug as length is checked before converting to UTF8
+            # which will cause cache_key to grow in size.
+            if len(cache_key) + len(namespace) > int(key_length):
+                cache_key = sha1(cache_key.encode('utf-8')).hexdigest()
+
             def go():
-                return func(*args)
+                return func(*args, **kwargs)
+            # save org function name
+            go.__name__ = '_cached_%s' % (func.__name__,)
 
             return cache[0].get_value(cache_key, createfunc=go)
         cached._arg_namespace = namespace
@@ -572,13 +605,11 @@ def _cache_decorate(deco_args, manager, kwargs, region):
         return cached
     return decorate
 
+
 def _cache_decorator_invalidate(cache, key_length, args):
     """Invalidate a cache key based on function arguments."""
 
-    try:
-        cache_key = " ".join(map(str, args))
-    except UnicodeEncodeError:
-        cache_key = " ".join(map(str, args))
+    cache_key = u_(" ").join(map(u_, args))
     if len(cache_key) + len(cache.namespace_name) > key_length:
-        cache_key = sha1(cache_key).hexdigest()
+        cache_key = sha1(cache_key.encode('utf-8')).hexdigest()
     cache.remove_value(cache_key)
